with Ada.Characters.Handling,
     Ada.Strings.Fixed,
     Ada.Strings.Maps;

package body Http_Urls is
  use Ada.Characters.Handling;
  use Ada.Strings.Fixed;
  use Ada.Strings.Maps;
  Part_Separators: constant Ada.Strings.Maps.Character_Set
    := Ada.Strings.Maps.To_Set(":/?#");
  Host_Enders: constant Ada.Strings.Maps.Character_Set
    := Ada.Strings.Maps.To_Set("/?#");
  Path_Enders: constant Ada.Strings.Maps.Character_Set
    := Ada.Strings.Maps.To_Set("?#");
  Slash   : constant Ada.Strings.Maps.Character_Set
    := Ada.Strings.Maps.To_Set("/");
  Fragment_Starter: constant Ada.Strings.Maps.Character_Set
    := Ada.Strings.Maps.To_Set("#");
  Numerals: constant Ada.Strings.Maps.Character_Set
    := Ada.Strings.Maps.To_Set("0123456789");

  function Scheme(Url     : String;
                  Base_Url: String := "") return String is
  -- return "http:" if no host part in either URL or Base_URL
    Si : Natural := Index(Url, Part_Separators);
  begin
    if Si /= 0 and then Url(Si) = ':' then
      return To_Lower(Url(Url'First .. Si));
    end if;
    if Base_Url'Length /= 0 then
      return Scheme(Base_Url, "");
    end if;
    return "http:";
  end Scheme;

  function Host(Url     : String;
                Base_Url: String := "") return String is
  -- return "" if neither URL nor Base_URL has a host
  -- NOTE: A legal http url must have a non-null host
    Si : Natural := Index(Url, Part_Separators);
    Di : Natural;
  begin
    if Si /= 0 and then Url(Si) = ':' then
      Si := Si + 1;
    else
      Si := Url'First;
    end if;
    if Si < Url'Last and then Url(Si) = '/' and then Url(Si + 1) = '/' then
      Di := Index(Url(Si + 2 .. Url'Last), Host_Enders);
      if Di = 0 then
        Di := Url'Last + 1;
      end if;
      return To_Lower(Url(Si .. Di - 1));
    end if;
    -- no //, try to get Host from Base_URL
    if Base_Url'Length /= 0 then
      return Host(Base_Url, "");
    end if;
    -- no // anywhere in sight
    return "";
  end Host;

  function Path(Url     : String;
                Base_Url: String := "") return String is
  -- result'length >= 1
  -- result(result'first) = '/'
    Si : Natural := Index(Url, Part_Separators);
    Di : Natural;

    function Compatible return Boolean is
      Bh      : constant String := Host(Base_Url);
      Uh      : constant String := Host(Url);
    begin
      return (Uh'Length = 0 or else Bh'Length = 0 or else Uh = Bh);
    end Compatible;
  begin
    if Url'Length = 0 then
      if Base_Url'Length = 0 then
        return "/";
      else
        return Path(Base_Url, "");
      end if;
    end if;
    if Si /= 0 and then Url(Si) = ':' then
      Si := Si + 1;
    else
      Si := Url'First;
    end if;
    if Si < Url'Last and then Url(Si) = '/' and then Url(Si + 1) = '/' then
      Si := Si + 2;
      Si := Index(Url(Si .. Url'Last), Host_Enders);
      if Si = 0 or else Url(Si) /= '/' then
        if Compatible and then Base_Url'Length > 1 then
          return Path(Base_Url, "/");
        else
          return "/";
        end if;
      end if;
      -- si is / at start of path
      Di := Index(Url(Si .. Url'Last), Path_Enders);
      if Di = 0 then
        Di := Url'Last + 1;
      end if;
      return Url(Si .. Di - 1);
    end if;
    -- no //
    if Si > Url'Last then
      if Compatible and then Base_Url'Length > 1 then
        return Path(Base_Url, "/");
      else
        return "/";
      end if;
    end if;
    Di := Index(Url(Si .. Url'Last), Path_Enders);
    if Di = Si then
      return Path(Base_Url, "");
    end if;
    if Di = 0 then
      Di := Url'Last + 1;
    end if;
    if Url(Si) = '/' then
      return Url(Si .. Di - 1);
    end if;
    if not Compatible then
      return "/" & Url(Si .. Di - 1);
    end if;
    declare
      Base_Path: constant String := Path(Base_Url, "");
      Bi      : constant Natural := Index(Base_Path, Slash, Going => Ada.
        Strings.Backward);
    begin
      if Bi = 0 then
        return "/" & Url(Si .. Di - 1);
      else
        return Base_Path(Base_Path'First .. Bi) & Url(Si .. Di - 1);
      end if;
    end;
  end Path;

  function Query(Url     : String;
                 Base_Url: String := "") return String is
  -- return "?whatever" or ""
    Si      : Natural := Index(Url, Path_Enders);
    Di      : Natural;
  begin
    if Si = 0 then
      if Base_Url'Length > 0 then
        return Query(Base_Url, "");
      else
        return "";
      end if;
    end if;
    if Url(Si) = '?' then
      Di := Index(Url, Fragment_Starter);
      if Di = 0 then
        Di := Url'Last + 1;
      end if;
      return Url(Si .. Di - 1);
    end if;
    return Query(Base_Url, "");
  end Query;

  function Fragment(Url     : String;
                    Base_Url: String := "") return String is
  -- return "#whatever" or ""
    Si      : Natural := Index(Url, Fragment_Starter);
  begin
    if Si = 0 then
      if Base_Url'Length > 0 then
        return Fragment(Base_Url, "");
      else
        return "";
      end if;
    end if;
    return Url(Si .. Url'Last);
  end Fragment;

  function Portless_Host(Host : String) return String is
  -- Strips off any trailing :nn
    Si : Natural := Index(Host, Numerals, Test => Ada.Strings.Outside,
                          Going => Ada.Strings.Backward);
  begin
    if Si /= 0 and then Host(Si) = ':' then
      return Host(Host'First .. Si - 1);
    else
      return Host;
    end if;
  end Portless_Host;

  function Port(Host : String) return Natural is
  -- Returns value of any trailing :nn, or zero
    Si : Natural := Index(Host, Numerals, Test => Ada.Strings.Outside,
                          Going => Ada.Strings.Backward);
  begin
    if Si /= 0 and then Host(Si) = ':' and then Si < Host'Last then
      return Natural'Value(Host(Si + 1 .. Host'Last));
    else
      return 0;
    end if;
  end Port;

end Http_Urls;
