with Ada.Calendar,
     Ada.Characters.Handling,
     Ada.Exceptions,
     Ada.Strings,
     Ada.Strings.Fixed,
     Ada.Text_Io,
     Http_Urls,
     Log,
     Statistics,
     Str_Sto,
     Str_Sto.Store;

package body Urls is

  Ignore_cgi: constant Boolean := True;

  function Squeeze_Path(Source : String) return String is
  -- squeeze a/b/c/../d/e  to a/b/d/e
  --     /./ to /
  --     //  to /
    Target  : String := Source;
    Si      : Natural := Source'First;
    Di      : Natural := Target'First;
  begin
    while Si <= Source'Last loop
      if Si + 3 <= Source'Last and then Source(Si .. Si + 3) = "/../" then
        while Di > Target'First loop
          Di := Di - 1;
          exit when Target(Di) = '/';
        end loop;
        Si := Si + 3;
      elsif Si + 2 <= Source'Last and then Source(Si .. Si + 2) = "/./" then
        Si := Si + 2;
      elsif Si + 1 <= Source'Last and then Source(Si .. Si + 1) = "//" then
        Si := Si + 1;
      else
        Target(Di) := Source(Si);
        Si := Si + 1;
        Di := Di + 1;
      end if;
    end loop;
    return Target(Target'First .. Di - 1);
  end Squeeze_Path;

  package Store is new Str_Sto.Store(Indices);

  Status_Code: array (Indices) of Status_Codes := (others => No_Response_Yet);
  Who_Said: array (Indices) of Counts := (others => 0);
  Content_Length: array (Indices) of Content_Lengths := (others => 0);

  package Content_Length_Io is new Ada.Text_Io.Integer_Io(Content_Lengths);

  package Int_Io is new Ada.Text_Io.Integer_Io(Integer);

  Last_Modified: array (Indices) of Ada.Calendar.Time := (others => Null_Time);

  Retry_List: array (Indices) of Counts := (others => 0);
  -- Procedure Start is out here doing stuff that takes a fair amount of
  -- stack space, then it calls Database.Start  That way the Database task
  -- as a (long running) whole doesn't need so much stack space.

  protected body Operational_Statistics is

    procedure Inc_Root_Domain_Open_Count is
    begin
      Root_Domain_Open_Count := Root_Domain_Open_Count + 1;
    end Inc_Root_Domain_Open_Count;

    procedure Inc_Fetch_Root_Page_Count is
    begin
      Fetch_Root_Page_Count := Fetch_Root_Page_Count + 1;
    end Inc_Fetch_Root_Page_Count;

    procedure Inc_Drain_Count is
    begin
      Drain_Count := Drain_Count + 1;
    end Inc_Drain_Count;

    procedure Inc_Drain_Byte_Count(N : in Long_Integer) is
    begin
      Drain_Byte_Count := Drain_Byte_Count + N;
    end Inc_Drain_Byte_Count;

    procedure Inc_Drain_Wait(T : in Duration) is
    begin
      Drain_Wait := Drain_Wait + T;
    end Inc_Drain_Wait;

    function Get_Root_Domain_Open_Count return Natural is
    begin
      return Root_Domain_Open_Count;
    end Get_Root_Domain_Open_Count;

    function Get_Fetch_Root_Page_Count return Natural is
    begin
      return Fetch_Root_Page_Count;
    end Get_Fetch_Root_Page_Count;

    function Get_Drain_Count return Natural is
    begin
      return Drain_Count;
    end Get_Drain_Count;

    function Get_Drain_Byte_Count return Long_Integer is
    begin
      return Drain_Byte_Count;
    end Get_Drain_Byte_Count;

    function Get_Drain_Wait return Duration is
    begin
      return Drain_Wait;
    end Get_Drain_Wait;

  end Operational_Statistics;

  procedure Initialize(Requested_Root_Url: in String) is
    use type Claw.Sockets.Network_Address_Type;
    Info    : Claw.Sockets.Host_Info_Type;
    Host    : constant String := Http_Urls.Host(Requested_Root_Url);
    use Ada.Strings.Fixed;
    Root_Url_Path: constant String := Http_Urls.Path(Requested_Root_Url, "");
    -- Path always starts with "/"
    Di      : Positive := Index(Root_Url_Path, "/", Ada.Strings.Backward);
  begin
    Root_Url_Host := new String'(Host);
    Root_Url_Domain_Ip_Address
      := Claw.Sockets.Read_As_Address(Host(Host'First + 2 .. Host'Last));
    if Root_Url_Domain_Ip_Address = 0 then
      Claw.Sockets.Lookup_Host_Info(Host(Host'First + 2 .. Host'Last), Info,
                                    Opening_Patience);
      Root_Url_Domain_Ip_Address := Claw.Sockets.Host_Address(Info, 1);
    end if;
    Root_Url_Has_Path := (Root_Url_Path'Length > 1);
    Root_Url_Filename_Length := 0;
    if Di /= Root_Url_Path'Last
    and then Index(Root_Url_Path(Di + 1 .. Root_Url_Path'Last), ".") /= 0 then
      Root_Url_Filename_Length := Root_Url_Path'Last - Di;
    end if;
    Statistics.Set_Name(Requested_Root_Url);
    declare
      use type Claw.Sockets.Network_Address_Type;
      use type Claw.Sockets.Network_Address_Bytes_Type;
      Root_Addr: Claw.Sockets.Network_Address_Bytes_Type
        := Claw.Sockets.Network_Address_Bytes_Of(Root_Url_Domain_Ip_Address);
      Proxy_Addr: Claw.Sockets.Network_Address_Bytes_Type
        := Claw.Sockets.Network_Address_Bytes_Of(Proxy_Ip);
    begin
      Root_Addr.C := 0;
      Proxy_Addr.C := 0; -- ignore .c  and .d
      Root_Addr.D := 0;
      Proxy_Addr.D := 0;
      Use_Proxy_For_Main := (Proxy_Ip /= 0 and Proxy_Addr /= Root_Addr);
    end;
    Root_Url := new String'(Requested_Root_Url); -- Root_URL stuff set up
  end Initialize;

  procedure Dump_Results(Result_File: in  Ada.Text_Io.File_Type) is
    use Ada.Text_Io;
    use type Ada.Calendar.Time;
    use type Store.Order_Positive;
    List : Store.Ordered_Lists := Store.Get_Ordered_Contents;
    -- watch for something with the same Content_Length,
    -- status_code, Last_Modified, and one character, a '/', longer
    Prev_Size: Content_Lengths := 0;
    Prev_Status: Status_Codes := Good_200_Response;
    Prev_Date: Ada.Calendar.Time := Null_Time;
    Prev_Length: Natural := 0;
    Dup_Count: Natural := 0;
    Dup_Size: Content_Lengths := 0;
  begin
    for I in List'range loop
      declare
        Url     : constant String := Store.Get(List(I));
      begin
        if Url'Length = Prev_Length + 1
        and then Url(Url'Last) = '/'
        and then Content_Length(List(I)) = Prev_Size
        and then Status_Code(List(I)) = Prev_Status
        and then Last_Modified(List(I)) = Prev_Date
        and then Url = Store.Get(List(I - 1)) & '/' then
          Put_Line(Result_File, "=/");
          Dup_Count := Dup_Count + 1;
          if Content_Length(List(I)) /= Size_Not_Available then
            Dup_Size := Dup_Size + Content_Length(List(I));
          end if;
          Prev_Length := Url'Length;
        else
          if Content_Length(List(I)) = Size_Not_Available then
            Put(Result_File, "    n.a.");
          else
            Content_Length_Io.Put(Result_File, Content_Length(List(I)),
                                  Width => 8);
          end if;
          Put(Result_File, ' ');
          if Last_Modified(List(I)) = Null_Time then
            Put(Result_File, "      n.a.");
          else
            Int_Io.Put(Result_File, Ada.Calendar.Month(Last_Modified(List(I))),
                       Width => 2);
            Put(Result_File, '/');
            Int_Io.Put(Result_File, Ada.Calendar.Day(Last_Modified(List(I))),
                       Width => 2);
            Put(Result_File, '/');
            Int_Io.Put(Result_File, Ada.Calendar.Year(Last_Modified(List(I))),
                       Width => 4);
          end if;
          Put(Result_File, ' ');
          Put(Result_File, Url);
          case Status_Code(List(I)) is
            when Good_200_Response | Response_Text_Html =>
              null;
            when Cant_Connect =>
              Put(Result_File, " *** Can't Connect");
            when Bad_Chunk_Read =>
              Put(Result_File, " *** Bad chunk terminated read");
            when No_Response_Yet =>
              Put(Result_File, " *** No Response");
            when Http_Codes =>
              Put(Result_File, " *** HTTP "
                               & Status_Codes'Image(Status_Code(List(I))));
            when Socket_Codes =>
              Put(Result_File, " *** "
                & Claw.Sockets.Image(Claw.Sockets.Error_Codes(Status_Code(List(I)))));
            when others =>
              Put(Result_File, " *** ??? "
                               & Status_Codes'Image(Status_Code(List(I))));
          end case;
          if Status_Code(List(I)) /= Good_200_Response
          and Status_Code(List(I)) /= Response_Text_Html
          and Who_Said(List(I)) /= 0 then
            Put(Result_File, " from " & Store.Get(Who_Said(List(I))));
          end if;
          New_Line(Result_File);
          Prev_Size := Content_Length(List(I));
          Prev_Status := Status_Code(List(I));
          Prev_Date := Last_Modified(List(I));
          Prev_Length := Url'Length;
        end if;
      end;
    end loop;
    Statistics.Print(Result_File => Result_File);
    if Dup_Count > 0 then
      Log.Note("NOTE:" & Natural'Image(Dup_Count) & " URLs, totaling"
        & Content_Lengths'Image(Dup_Size)
        & " bytes, appear double counted, with/without a trailing '/'");
    end if;
    Log.Note(Natural'Image(Operational_Statistics.Get_Root_Domain_Open_Count)
      & " opens in"
      & Natural'Image(Operational_Statistics.Get_Fetch_Root_Page_Count)
      & " fetches");
    if Operational_Statistics.Get_Drain_Count > 0 then
      Log.Note(Natural'Image(Operational_Statistics.Get_Drain_Count)
        & " Drained,"
        & Long_Integer'Image(Operational_Statistics.Get_Drain_Byte_Count)
        & " bytes,"
        & Duration'Image(Operational_Statistics.Get_Drain_Wait) & " seconds");
    end if;
  end Dump_Results;

  task body Database_Task is
    use Ada.Characters.Handling;
    use type Ada.Calendar.Time;
    Next_To_Do: Indices := 1;
    Url_Count: Counts := 0;
    In_Process: Natural := 0;
    New_Url : Url_Ptr_Type;
    Next_Work_May_Be_Root: Boolean := True;
    -- Retry handling
    Next_Retry: Counts := 0; -- next entry in Retry_List, or zero
    Last_Retry: Counts := 0; -- last entry in Retry_List, or zero

    procedure Add_Retry(Id : in Indices) is
    begin
      --    if Next_Retry = 0 then
      --      Next_Retry := id;
      --    else
      --      Retry_List(Last_Retry) := id;
      --    end if;
      --    Last_Retry := id;
      -- make it LIFO instead of FIFO
      -- so retries won't occur in the same order as initial tries.
      if Next_Retry = 0 then
        Next_Retry := Id;
      else
        Retry_List(Id) := Next_Retry;
        Next_Retry := Id;
      end if;
    end Add_Retry;

    procedure Remove_Retry is
    begin
      Next_Retry := Retry_List(Next_Retry);
      -- fails if Next_Retry = 0
    end Remove_Retry;

  begin

    select
      accept Start(Requested_Root_Url: in String) do
        Initialize(Requested_Root_Url);
      end Start;
    or
      terminate;
    end select;

    Url_Count := Store.Put(Http_Urls.Scheme(Root_Url.all)
                           & Root_Url_Host.all
                           & Http_Urls.Path(Root_Url.all));
    while Next_To_Do <= Url_Count or In_Process > 0 or Next_Retry /= 0 loop
      Free(New_Url); -- get rid of any old one
      exit when In_Process = 0 and Statistics.Kill_Request;
      if Next_To_Do <= Url_Count then
        Next_Work_May_Be_Root := Http_Urls.Host(Store.Get(Next_To_Do)) =
          Root_Url_Host.all;
      else
        Next_Work_May_Be_Root := True;
      end if;

      select
        -- If In_Process = 0 but there's still work to do, then we're
        -- clearly not overloading Root so we will assign root work
        -- even if asked for non-root.  Thus even if all workers
        -- prefer non-root, we'll still assign root work and eventually
        -- get it done.
        when (Next_To_Do <= Url_Count or Next_Retry /= 0)
             and (In_Process = 0 or not Next_Work_May_Be_Root) =>
        accept Get_Non_Root_Work(Url_To_Do_Index: out Counts;
                                 Url     :    out Url_Ptr_Type;
                                 Second_Try:  out Boolean) do
          if Statistics.Kill_Request then
            -- after op kill request
            Url_To_Do_Index := 0; -- don't start anything new
          elsif Next_To_Do <= Url_Count then
            Url_To_Do_Index := Next_To_Do;
            Url := new String'(Store.Get(Next_To_Do));
            In_Process := In_Process + 1;
            Second_Try := False;
            Next_To_Do := Next_To_Do + 1;
          elsif Next_Retry /= 0 then
            Url_To_Do_Index := Next_Retry;
            Url := new String'(Store.Get(Next_Retry));
            In_Process := In_Process + 1;
            Second_Try := True;
            Remove_Retry;
          end if;
        end Get_Non_Root_Work;
      or
        when Next_To_Do <= Url_Count or Next_Retry /= 0 =>
        accept Get_Any_Work(Url_To_Do_Index: out Counts;
                            Url     :    out Url_Ptr_Type;
                            Second_Try:  out Boolean) do
          if Statistics.Kill_Request then
            -- after op kill request
            Url_To_Do_Index := 0; -- don't start anything new
          elsif Next_To_Do <= Url_Count then
            Url_To_Do_Index := Next_To_Do;
            Url := new String'(Store.Get(Next_To_Do));
            In_Process := In_Process + 1;
            Second_Try := False;
            Next_To_Do := Next_To_Do + 1;
          elsif Next_Retry /= 0 then
            Url_To_Do_Index := Next_Retry;
            Url := new String'(Store.Get(Next_Retry));
            In_Process := In_Process + 1;
            Second_Try := True;
            Remove_Retry;
          end if;
        end Get_Any_Work;
      or
        accept Note_New_Url(S       : in     String; -- raw URL from file
                            Source_Url: in   String;
                              -- where we found this URL, for debugging
                            Base    : in     String;
                            Referrer: in     Indices) do
          -- index of Source_URL
          -- clean S, using Base and reporting errors with Source_URL
          -- If S is then good, note_new_url it to store for processing
          Statistics.Increment_Links_Scanned;
          declare
            Scheme  : constant String := Http_Urls.Scheme(S, Base);
            Host    : constant String := Http_Urls.Host(S, Base);
            Path    : constant String := Squeeze_Path(Http_Urls.Path(S, Base));
          begin
            --&&Ugh, patch to chase links broken only because they use the old machine name:
            if Http_Urls.Portless_Host(Host) = "//sw-eng.falls-church.va.us" then
              Statistics.Increment_Old_Domain_Links;
              Log.Note(Source_Url & " Remove old domain:" & S);
              New_Url := new String'(Scheme & Root_Url_Host.all & Path);
            elsif Http_Urls.Query(S, Base)'Length = 0 then
              -- (ignore cgi queries)
              New_Url := new String'(Scheme & Host & Path);
            end if;
            if New_Url /= null then
              for I in New_Url.all'range loop
                if Is_Control(New_Url(I)) then
                  Log.Note(Source_Url
                           & "*Bad URL* Ctrl char at"
                           & Natural'Image(I - New_Url.all'First + 1)
                           & " text:" & New_Url.all);
                  Statistics.Increment_Bad_Links;
                  Free(New_Url);
                  exit;
                elsif Ignore_cgi and then To_Lower(New_Url(I)) = 'c'
                and then I + 2 <= New_Url.all'Last
                and then To_Lower(New_Url(I .. I + 2)) = "cgi"
                and then (I + 3 > New_Url.all'Last
                          or else not Is_Letter(New_Url(I + 3)))
                and then (I = New_Url.all'First
                          or else not Is_Letter(New_Url(I - 1))) then
                  Free(New_Url); -- This is a cgi URL, ignore it.
                  exit;
                end if;
              end loop;
            end if;
            if New_Url /= null
            and then (Scheme = "http:" or Scheme = "ftp:") then
              if not Store.Is_Present(New_Url.all) then
                Url_Count := Store.Put(New_Url.all); -- save good URL!
                Who_Said(Url_Count) := Referrer;
              end if;
            end if;
            Free(New_Url);
          end;
        end Note_New_Url;
      or
        accept Post_Work_Result(Id      : in     Indices;
                                Size    : in     Content_Lengths;
                                Date    : in     Ada.Calendar.Time;
                                Status  : in     Status_Codes;
                                Foreign : in     Boolean;
                                Failed  : in     Boolean;
                                Scanned : in     Boolean;
                                Url_String: in   String;
                                Second_Try: in   Boolean) do
          -- Statistics info: Size is the size of item that the processed URI
          -- points at; Foreign is true if the item is not on the processed
          -- site; Failed is true if the item failed to process properly;
          -- Scanned is true if the text was scanned; Second_Try is true
          -- if this is the second try to access a URL.
          In_Process := In_Process - 1;
          if not Second_Try then
            Content_Length(Id) := Size;
            Last_Modified(Id) := Date;
            Status_Code(Id) := Status;
          else
            -- only replace if first shot had default value
            if Content_Length(Id) = Size_Not_Available then
              Content_Length(Id) := Size;
            end if;
            if Last_Modified(Id) = Null_Time then
              Last_Modified(Id) := Date;
            end if;
          end if;
          Status_Code(Id) := Status; -- most recent Status
          if Failed and (not Second_Try) then
            -- Note_new_url to retry list
            Add_Retry(Id);
          end if;
          Statistics.Finished_One(Claw.Dword(Content_Lengths'Max(0, Size)),
                                  Foreign, Failed, Scanned, Url_String,
                                  Second_Try, Claw.Dword(Url_Count));
        end Post_Work_Result;
      end select;

    end loop;

    -- In_Process = 0 so no workers are busy and they should all be
    -- attempting to get work.  We'll give them work with
    -- URL_To_Do_Index=0, ie, "quit", until no more ask.
    loop
      select
        accept Get_Any_Work(Url_To_Do_Index: out Counts;
                            Url     :    out Url_Ptr_Type;
                            Second_Try:  out Boolean) do
          Url_To_Do_Index := 0;
        end Get_Any_Work;
      or
        accept Get_Non_Root_Work(Url_To_Do_Index: out Counts;
                                 Url     :    out Url_Ptr_Type;
                                 Second_Try:  out Boolean) do
          Url_To_Do_Index := 0;
        end Get_Non_Root_Work;
      else
        exit;
      end select;
    end loop;

    -- Now we're ready to dump results when asked.
    accept Dump(Result_File: in  Ada.Text_Io.File_Type) do
      Dump_Results(Result_File);
    end Dump;

    if Next_To_Do <= Url_Count then
      Log.Note("----- Killed with at least"
        & Counts'Image(Url_Count - Next_To_Do + 1) & " remaining");
    end if;

  exception
    when Oops : others =>
      Log.Note("oops:" & Counts'Image(Next_To_Do) & Counts'Image(Url_Count)
        & Natural'Image(In_Process));
      Log.Note(Ada.Exceptions.Exception_Information(Oops));
  end Database_Task;

  type Port_Exception_Counts is range 0 .. 50;
  subtype Port_Exception_Indices
    is Port_Exception_Counts range 1 .. Port_Exception_Counts'Last;

  package Port_Exception_Host_Store is new Str_Sto.Store(Port_Exception_Indices);

  Port_Exception_Port: array (Port_Exception_Indices) of Claw.Sockets.Port_Type
    := (others => 0);

  -- we want lookup serialized so cacheing in Port_Exception_Host_Store
  -- doesn't get confused.
  protected Port_Exception_Lookup is
    function Exception_Port(Host    : in     String) return Claw.Sockets.Port_Type;
  end Port_Exception_Lookup;

  protected body Port_Exception_Lookup is
    function Exception_Port(Host : in String) return Claw.Sockets.Port_Type is
    begin
      if Port_Exception_Host_Store.Is_Present(Host) then
        return Port_Exception_Port(Port_Exception_Host_Store.Put(Host));
      else
        return 0;
      end if;
    end Exception_Port;
  end Port_Exception_Lookup;

  function Exception_Port(Host : in String) return Claw.Sockets.Port_Type is
  begin
    return Port_Exception_Lookup.Exception_Port(Host);
  end Exception_Port;

  procedure Create_Port_Exception_List(PE_File_Name : in String) is
    use Ada.Text_IO;
    use type Claw.Sockets.Port_Type;
    PE_File : File_Type;
    Line : String(1 .. 100);
    Last : Natural;
    Port : Claw.Sockets.Port_Type;
    Exception_Index : Port_Exception_Indices;
  begin
    Open(PE_File, Ada.Text_IO.In_File, PE_File_Name);
    loop
      Get_Line(PE_File, Line, Last);
      Port := Claw.Sockets.Port_Type(HTTP_URLs.Port(Line(1 .. Last)));
      if Port = 0 then
        Log.Note("no port in " & Line(1 .. Last));
      else
        Exception_Index := Port_Exception_Host_Store.Put
          (HTTP_URLs.Portless_Host(Line(1 .. Last)));
        Port_Exception_Port(Exception_Index) := Port;
      end if;
    end loop;
  exception
    when others =>
      if Is_Open(PE_File) then
        Close(PE_File);
      end if;
  end Create_Port_Exception_List;

end URLs;
