with Ada.Calendar,
     Ada.Text_IO,
     Ada.Unchecked_Deallocation,
     Claw,
     Claw.Directories,
     Claw.Sockets;
package URLs is

  Opening_Patience : constant Duration := 0.0; -- infinitely patient

  type Counts is range 0 .. 25_000;     -- 15000
  subtype Indices is Counts range 1 .. Counts'last;

  type Content_Lengths is new Long_Integer;
  Size_Not_Available : constant Content_Lengths := -1; -- flag value

  Null_Time : constant Ada.Calendar.Time
    := Claw.Directories.Unknown_Time;  -- 1/1/1980 00:00:00

    type URL_Ptr_Type is access String;
    procedure Free is new Ada.Unchecked_Deallocation(String, URL_Ptr_Type);

  type Status_Codes is new Claw.Int;
  -- save HTTP nnn code, or Claw.Sockets.Error_Codes, or our own codes
  -- Windows sometimes returns wild codes (0 and -32764 have both appeared)
  Cant_Connect : constant Status_Codes := 1;
  Bad_Chunk_Read : constant Status_Codes := 2;
  No_Response_Yet : constant Status_Codes := 3;  -- state after Open_Socket
  Good_200_Response : constant Status_Codes := 4; -- after HTTP/1.n 200 xxx
  Response_Text_HTML : constant Status_Codes := 5; -- Content-type: text/html
  subtype HTTP_Codes is Status_Codes range 100 .. 999;  -- 404 et al
  subtype Socket_Codes is Status_Codes range 10000 .. Status_Codes'last;

  function Squeeze_Path(Source : String) return String;
  -- squeeze a/b/c/../d/e  to a/b/d/e
  --     /./ to /
  --     //  to /

  task type Database_Task is  -- make it a task type so we can 'size it

    entry Start(Requested_Root_URL : in String); -- start by Adding Root_Url

    entry Get_Any_Work(URL_To_Do_Index : out Counts;
                       URL : out URL_Ptr_Type;
                       Second_Try : out Boolean);
    -- URL_To_Do_Index = 0 if there is no work, and no prospect of new work

    entry Get_Non_Root_Work(URL_To_Do_Index : out Counts;
                            URL : out URL_Ptr_Type;
                            Second_Try : out Boolean);
    -- Try to get non-root work if possible, so as not to overload
    -- Root's server.  If necessary, however, this will assign root work.
    -- URL_To_Do_Index = 0 if there is no work, and no prospect of new work

    entry Note_new_url(S : in String;  -- raw URL from file
              Source_URL : in String;  -- referring URL
              Base : in String;        -- current "path"
              referrer : in Indices);  -- referring URL index
    -- clean S, using Base and reporting errors with Source_URL
    -- If S is then good, note_new_url it to store for processing

    entry Post_work_result (ID   : in indices; -- finished processing some url
                        Size : in Content_Lengths;
                        Date : in Ada.Calendar.Time;
                        Status : in Status_Codes;
                        Foreign : in Boolean;
                        Failed : in Boolean;
                        Scanned : in Boolean;
                        URL_String : in String;
                        Second_Try : in Boolean);
    -- save statistics

    entry Dump(Result_File : in Ada.Text_IO.File_Type);
    -- Dump is accepted when there is no more work to do or in process

  end Database_Task;
  for Database_Task'Storage_Size use 24_671;

  Database : Database_Task;

  protected Operational_Statistics is
    procedure Inc_Root_Domain_Open_Count;
    procedure Inc_Fetch_Root_Page_Count;
    procedure Inc_Drain_Count;
    procedure Inc_Drain_Byte_Count(N : in Long_Integer);
    procedure Inc_Drain_Wait(T : in Duration);
    function Get_Root_Domain_Open_Count return Natural;
    function Get_Fetch_Root_Page_Count return Natural;
    function Get_Drain_Count return Natural;
    function Get_Drain_Byte_Count return Long_Integer;
    function Get_Drain_Wait return Duration;
  private
    Root_Domain_Open_Count : Natural := 0;
    Fetch_Root_Page_Count : Natural := 0;
    Drain_Count : Natural := 0;
    Drain_Byte_Count : Long_Integer := 0;
    Drain_Wait : Duration := 0.0;
  end Operational_Statistics;

  Proxy_IP : Claw.Sockets.Network_Address_Type := 0;  -- 0 => no proxying
--      Claw.Sockets.Network_Address_Of( (a=>1, b=> 2, c=> 3, d=> 4) );
        -- Do we need to use a proxy for accessing external HTTP?
        -- (Not on the local ("1.2.xx.xx") network.)

  -- These have their values filled in by Database.Start - don't use
  -- before Root_URL /= null, and don't change them.

  Root_URL_Domain_IP_Address : Claw.Sockets.Network_Address_Type;
  Root_URL_Has_Path : Boolean;
  Root_URL_Filename_Length : Natural;
  Use_Proxy_for_Main : Boolean;

  Root_URL_Host : URL_Ptr_Type;
  Root_URL : URL_Ptr_Type; -- when this is non-zero (at the end of Start),
  -- the above have legitimate values.

  procedure Create_Port_Exception_List(PE_File_Name : in String);

  function Exception_Port(Host : in String) return Claw.Sockets.Port_Type;
  -- Must not be called simultaneously with Create_Port_Exception_List.

end URLs;
