with Ada.Calendar,
     Ada.Characters.Handling,
     Ada.Exceptions,
     Ada.Strings,
     Ada.Strings.Maps,
     Claw,
     Claw.Sockets,
     Html_Scan,
     Http_Urls,
     Log,
     Str_Sto,
     Str_Sto.Store,
     Te_Text_Line_Sockets,
     Urls;

package body Workers is

  Http_Proxy_Port: Claw.Sockets.Port_Type := 8080;
  Ftp_Proxy_Port: Claw.Sockets.Port_Type := 8021;

  Lower_Case: constant Ada.Strings.Maps.Character_Mapping := Ada.Strings.Maps.
    To_Mapping(From => "ABCDEFGHIJKLMNOPQRSTUVWXYZ",
                 To => "abcdefghijklmnopqrstuvwxyz");
  Non_Filename_Chars: constant Ada.Strings.Maps.Character_Set
    := Ada.Strings.Maps.To_Set("/#?:");

  -- cache some domains to save some time
  type Domain_Counts is range 0 .. 1000;
  subtype Domain_Indices is Domain_Counts range 1 .. Domain_Counts'Last;

  package Domain_Store is new Str_Sto.Store(Domain_Indices);
  Domain_Ip: array (Domain_Indices) of Claw.Sockets.Network_Address_Type;

  protected Domain_Ip_Cache is

    function Find(Name : String) return Claw.Sockets.Network_Address_Type;
    -- returns 0 if not found

    procedure Save(Name : in String;
                   Ip   : in Claw.Sockets.Network_Address_Type);
  end Domain_Ip_Cache;

  protected body Domain_Ip_Cache is

    function Find(Name : String) return Claw.Sockets.Network_Address_Type is
    -- returns 0 if not found
    begin
      if Domain_Store.Is_Present(Name) then
        return Domain_Ip(Domain_Store.Put(Name));
      else
        return 0;
      end if;
    end Find;

    procedure Save(Name : in String;
                   Ip   : in Claw.Sockets.Network_Address_Type) is
    begin
      Domain_Ip(Domain_Store.Put(Name)) := Ip;
    exception
      when Str_Sto.Overflow =>
        null; -- cache is full, sorry
    end Save;
  end Domain_Ip_Cache;

  procedure Find_And_Open(Socket  : in out Te_Text_Line_Sockets.Socket_Type;
                          Name    : in     String;
                          Port    : in     Claw.Sockets.Port_Type;
                          Timeout : in     Duration;
                          Result_Code: out Claw.Sockets.Error_Codes) is
    use type Claw.Sockets.Network_Address_Type;
    Ip      : Claw.Sockets.Network_Address_Type;
    Info    : Claw.Sockets.Host_Info_Type;
  begin
    Result_Code := 0;
    Ip := Domain_Ip_Cache.Find(Name);
    if Ip = 0 then
      begin
        Claw.Sockets.Lookup_Host_Info(Name, Info, Timeout);
      exception
        when Claw.Windows_Error =>
          Claw.Sockets.Get_Error_Code(Info, Result_Code);
          return;
      end;
      Ip := Claw.Sockets.Host_Address(Info, 1);
      Domain_Ip_Cache.Save(Name, Ip);
    end if;
    Te_Text_Line_Sockets.Open(Socket, Ip, Port, Timeout);
  exception
    when Claw.Windows_Error =>
      Te_Text_Line_Sockets.Get_Error_Code(Socket, Result_Code);
  end Find_And_Open;

  Mon : constant array (Ada.Calendar.Month_Number) of String(1 .. 3)
    := ("Jan", "Feb", "Mar", "Apr", "May", "Jun",
        "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");

  function Full_Date(Source : String) return Ada.Calendar.Time is
  -- scan an RFC 1123   Sun, 06 Nov 1994 08:49:37 GMT
  -- or 850(obsolete)   Sunday, 06-Nov-94 08:49:37 GMT
  -- Full Date and return as Ada.Calendar.Time
  -- raises Constraint_Error if problems
    use Ada.Calendar;
    Y2_Break: constant Integer := 70; -- -69 ->2069, -70 ->1970
    S       : constant String(1 .. Source'Length) := Source;
    Si      : Natural := 0;
    D       : Day_Number := Day_Number'First;
    M       : Month_Number := Month_Number'First;
    Y4      : Year_Number := Year_Number'First;
    Y2      : Integer range 0 .. 99 := 0;
    Hh      : Integer range 0 .. 23 := 0;
    Mm      : Integer range 0 .. 59 := 0;
    Ss      : Integer range 0 .. 59 := 0;
  begin
    if S(4) = ',' then
      -- Sun, 06 Nov 1994 08:49:37 GMT
      D := Ada.Calendar.Day_Number'Value(S(6 .. 7));
      for I in Month_Number loop
        if Mon(I) = S(9 .. 11) then
          M := I;
          exit;
        end if;
        if I = Month_Number'Last then
          raise Constraint_Error;
        end if;
      end loop;
      Y4 := Ada.Calendar.Year_Number'Value(S(13 .. 16));
      Hh := Integer'Value(S(18 .. 19));
      Mm := Integer'Value(S(21 .. 22));
      Ss := Integer'Value(S(24 .. 25));
    else
      -- Sunday, 06-Nov-94 08:49:37 GMT
      for I in S'range loop
        if S(I) = ' ' then
          Si := I + 1;
          exit;
        end if;
      end loop;
      D := Ada.Calendar.Day_Number'Value(S(Si .. Si + 1));
      for I in Ada.Calendar.Month_Number loop
        if Mon(I) = S(Si + 3 .. Si + 5) then
          M := I;
          exit;
        end if;
      end loop;
      Y2 := Ada.Calendar.Year_Number'Value(S(Si + 7 .. Si + 8));
      if Y2 < Y2_Break then
        Y4 := 2000 + Y2;
      else
        Y4 := 1900 + Y2;
      end if;
      Hh := Integer'Value(S(Si + 10 .. Si + 11));
      Mm := Integer'Value(S(Si + 13 .. Si + 14));
      Ss := Integer'Value(S(Si + 16 .. Si + 17));
    end if;
    return Time_Of(Y4, M, D, Day_Duration(Hh) * 3600
                             + Day_Duration(Mm) * 60
                             + Day_Duration(Ss));
  end Full_Date;

  procedure Fetch_Page(Me      : in     Character;
                       Is_Second_Try: in Boolean;
                       Url_Id  : in     Urls.Indices;
                       Scheme  : in     String;
                       Host    : in     String;
                       Path    : in     String;
                       Socket  : in out Te_Text_Line_Sockets.Socket_Type;
                       Is_Root_Domain: in Boolean;
                       Root_Connrefused: in out Boolean) is
    use type Urls.Content_Lengths;
    use type Urls.Status_Codes;
    use Ada.Characters.Handling;
    Page_Full_Url: constant String := Scheme & Host & Path;
    Use_Proxy: Boolean := False;
    Debug_Flag: Boolean := False;
    Must_Scan: Boolean := Is_Root_Domain; -- till shown otherwise
    Go_Direct_To_Get: Boolean := False; -- if *.htm[l] don't bother with HEAD
    Was_Foreign: Boolean := not Is_Root_Domain; -- or outside specified path
    Dont_Retry: Boolean := False; --  till shown otherwise
    Line    : String(1 .. 300); -- only for header so needn't be so big
    Last    : Natural;
    Result_Class: Character; -- Class of command result.
    Err     : Claw.Sockets.Error_Codes;
    Chunked_Transfer: Boolean := False;
    Close_After: Boolean := False;
    Item_Size: Urls.Content_Lengths := Urls.Size_Not_Available;
    Scanned_Item_Size: Urls.Content_Lengths := 0;
    Claimed_Item_Size: Urls.Content_Lengths := 0;
    Content_Length_Present: Boolean := False;
    Content_Length_Remaining: Urls.Content_Lengths := 0;
    Date    : Ada.Calendar.Time := Urls.Null_Time;
    Status  : Urls.Status_Codes := Urls.No_Response_Yet;
    -- error logging
    Source_Displayed: Boolean := False;

    procedure Note(S : in String) is
    begin
      if not Source_Displayed then
        Log.Note(Me & " Process:" & Page_Full_Url);
        Source_Displayed := True; -- done with it
      end if;
      Log.Note(Me & ' ' & S);
    end Note;
    -- scanning html file

    procedure Get_Line(Line    :    out String;
                       Last    :    out Natural) is
      use type Urls.Content_Lengths;
      Max_To_Read: Urls.Content_Lengths := Line'Length;
      -- Last < Line'first means we're out of data.  Could be
      -- Content_Length_Remaining = 0, Claw.Not_Found_Error (disconnect),
      -- or Te_Text_Line_Sockets.Get_Line timed out or got a zero chunk.
    begin
      if Content_Length_Present and Content_Length_Remaining = 0 then
        Last := Line'First - 1;
        return;
      end if;
      if Content_Length_Present
      and then Content_Length_Remaining < Max_To_Read then
        Max_To_Read := Content_Length_Remaining;
      end if;
      Te_Text_Line_Sockets.Get_Line(Socket,
        Line(Line'First .. Line'First + Integer(Max_To_Read) - 1),
        Last);
      Scanned_Item_Size
        := Scanned_Item_Size + Urls.Content_Lengths(Last - Line'First + 1);
      if Content_Length_Present then
        Content_Length_Remaining
          := Content_Length_Remaining
             - Urls.Content_Lengths(Last - Line'First + 1);
      end if;
    exception
      when Claw.Not_Found_Error =>
        Te_Text_Line_Sockets.Close(Socket);
        Last := Line'First - 1;
    end Get_Line;

    procedure Put_Ref(Url     : in     String; -- may be malformed
                      Base    : in     String) is
    begin
      Urls.Database.Note_New_Url(S => Url,
                                 Source_Url => Page_Full_Url,
                                 Base => Base,
                                 Referrer => Url_Id);
    end Put_Ref;

    procedure Get_Links is new Html_Scan.Get_Links(Socket.Max_Line_Length,
                                                   Get_Line, Put_Ref, Note);

    procedure Open_Socket is
      use type Claw.Sockets.Port_Type;
      use type Claw.Sockets.Network_Address_Type;
      use type Claw.Sockets.Error_Codes;
      Pure_Host: constant String -- drop initial //, any final :80
        := Http_Urls.Portless_Host(Host(Host'First + 2 .. Host'Last));
      Port    : Claw.Sockets.Port_Type;
    begin
      begin
        Port := Claw.Sockets.Port_Type(Http_Urls.Port(Host));
      exception
        when Bad_Port: others =>
          Note(Ada.Exceptions.Exception_Information(Bad_Port));
          Status := Urls.Cant_Connect;
          return;
      end;
      Use_Proxy := False;
      if Port = 0 then
        Port := Urls.Exception_Port(Pure_Host); -- check for exception port
      end if;
      if Port = 0 then
        -- check for proxy stuff
        if Urls.Proxy_Ip /= 0
        and then (Was_Foreign or else Urls.Use_Proxy_For_Main) then
          --          if Scheme = "ftp:" then
          --            Port := FTP_Proxy_Port;
          --          else
          Port := Http_Proxy_Port;
          --          end if;
          Use_Proxy := True;
        else
          --          if Scheme = "ftp:" then
          --            Port := Claw.Sockets.FTP_Port;
          --          else
          Port := Claw.Sockets.Http_Port;
          --          end if;
        end if;
      end if;
      -- should we use http port even if scheme name is ftp: ????
      if Use_Proxy then
        Te_Text_Line_Sockets.Open(Socket, Urls.Proxy_Ip, Port,
                                  Timeout => Urls.Opening_Patience);
        if Is_Root_Domain then
          Urls.Operational_Statistics.Inc_Root_Domain_Open_Count;
        end if;
      elsif not Is_Root_Domain then
        if Claw.Sockets.Read_As_Address(Pure_Host) = 0 then
          -- likely case
          begin
            Find_And_Open(Socket, Pure_Host, Port,
                          Timeout => Urls.Opening_Patience,
                          Result_Code => Err);
            case Err is
              when 0 =>
                null;
              when Claw.Sockets.Wsa_Host_Not_Found
                 | Claw.Sockets.Wsae_Timedout =>
                Status := Urls.Cant_Connect;
                return;
              when Claw.Sockets.Wsae_Connrefused =>
                if Is_Root_Domain and not Root_Connrefused then
                  Note("Connrefused");
                  Root_Connrefused := True;
                end if;
                Status := Urls.Cant_Connect;
                return;
              when others =>
                Status := Urls.Status_Codes(Err);
                return;
            end case;
          end;
        else
          Te_Text_Line_Sockets.Open(Socket,
                                    Claw.Sockets.Read_As_Address(Pure_Host),
                                    Port,
                                    Timeout => Urls.Opening_Patience);
        end if;
      else
        Urls.Operational_Statistics.Inc_Root_Domain_Open_Count;
        Te_Text_Line_Sockets.Open(Socket,
                                  Urls.Root_Url_Domain_Ip_Address,
                                  Port,
                                  Timeout => Urls.Opening_Patience);
      end if;
      Status := Urls.No_Response_Yet;
      -- A little delay seems needed to keep servers happy
      if Is_Second_Try then
        delay 0.5;
      else
        delay 0.1;
      end if;
    exception
      when Nf : Claw.Not_Found_Error =>
        -- unopenable domain
        Note("nf:" & Ada.Exceptions.Exception_Information(Nf));
        Status := Urls.Cant_Connect;
      when We : Claw.Windows_Error =>
        Te_Text_Line_Sockets.Get_Error_Code(Socket, Err);
        case Err is
          when Claw.Sockets.Wsa_Host_Not_Found | Claw.Sockets.Wsae_Timedout =>
            Status := Urls.Cant_Connect;
          when Claw.Sockets.Wsae_Connrefused =>
            if Is_Root_Domain and not Root_Connrefused then
              Note("Connrefused");
              Root_Connrefused := True;
            end if;
            Status := Urls.Cant_Connect;
          when others =>
            Status := Urls.Status_Codes(Err);
            Note("we" & Claw.Sockets.Image(Err) & " "
                 & Ada.Exceptions.Exception_Information(We));
        end case;
    end Open_Socket;

    procedure Scan_Header is
      use type Ada.Calendar.Time;
      Worth_Noting: Boolean := False;
      Seen_An_Info_Line: Boolean := False;
      First_Line_Patience: constant Duration := 60.0; -- till Seen_An_Info_Line
      Patience: constant Duration := 30.0;
      T       : Ada.Calendar.Time;
    begin
      loop
        -- scan returned header
        T := Ada.Calendar.Clock;
        begin
          if Seen_An_Info_Line then
            Te_Text_Line_Sockets.Get_Line(Socket, Line, Last,
                                          Timeout => Patience);
          else
            Te_Text_Line_Sockets.Get_Line(Socket, Line, Last,
                                          Timeout => First_Line_Patience);
          end if;
        exception
          when Claw.Not_Found_Error =>
            Te_Text_Line_Sockets.Close(Socket);
            Must_Scan := False; -- can't read entire Header
            return;
        end;
        if Ada.Calendar.Clock - T > 20.0 then
          Note("slow " & Duration'Image(Ada.Calendar.Clock - T)
               & Integer'Image(Last) & "|" & Line(1 .. Last));
        end if;
        if Last = 0 then
          -- timeout
          if Status /= Urls.Response_Text_Html then
            Must_Scan := False;
          end if;
          return;
        end if;
        -- drop trailing [cr|lf]*lf
        if Last > 0 and then Line(Last) = Ascii.Lf then
          for I in reverse 1 .. Last loop
            if Line(I) = Ascii.Cr or Line(I) = Ascii.Lf then
              Last := I - 1;
            else
              exit;
            end if;
          end loop;
        end if;
        if Last = 0 then
          -- empty line
          if Seen_An_Info_Line then
            if Status /= Urls.Response_Text_Html then
              Must_Scan := False;
            end if;
            return;
          else
            Note("empty line start");
          end if;
        end if;
        Seen_An_Info_Line := True;
        if Last = 1 and Line(1) = ' ' then
          -- illegal HTTP
          Note("single blank");
        end if;
        -- handle different header response lines
        if Last >= 12 and then Line(1 .. 5) = "HTTP/" then
          if Line(6 .. 8) = "1.0" then
            Close_After := True;
          end if;
          Result_Class := Line(10);
          if Line(9 .. 12) = " 200" then
            if Status = Urls.No_Response_Yet then
              Status := Urls.Good_200_Response;
            end if;
          else
            Status := Urls.Status_Codes'Value(Line(9 .. 12));
            Worth_Noting := True;
          end if;
          -- regardless of Result_Class, we want to see Location or Set-Cookie
        elsif Last > 10 and then Line(1 .. 10) = "Location: " then
          Urls.Database.Note_New_Url(Line(11 .. Last),
            Source_Url => Page_Full_Url, Base => Page_Full_Url,
            Referrer => Url_Id);
          Worth_Noting := True;
        elsif Last >= 12 and then Line(1 .. 12) = "Set-Cookie: " then
          Worth_Noting := True;
          -- for other info, pay attention to result class
        elsif Result_Class = '1' then
          -- informational only
          Worth_Noting := True;
        elsif Result_Class = '2' then
          -- success, scan header content
          if Last >= 26
          and then To_Lower(Line(1 .. 26)) = "transfer-encoding: chunked" then
            Chunked_Transfer := True;
          elsif Last >= 17
          and then To_Lower(Line(1 .. 17)) = "connection: close" then
            Close_After := True;
          elsif Last > 15
          and then To_Lower(Line(1 .. 15)) = "content-length:" then
            begin
              Claimed_Item_Size := Urls.Content_Lengths'Value(Line(16 .. Last));
              Item_Size := Claimed_Item_Size; -- unless proved otherwise
              Content_Length_Present := True;
              Content_Length_Remaining := Claimed_Item_Size;
              if not Must_Scan then
                Dont_Retry := True; -- we got the main thing
              end if;
            exception
              when Constraint_Error =>
                Worth_Noting := True;
            end;
          elsif Last > 15
          and then To_Lower(Line(1 .. 15)) = "last-modified: " then
            begin
              Date := Full_Date(Line(16 .. Last));
            exception
              when Constraint_Error =>
                Date := Urls.Null_Time;
                Worth_Noting := True;
            end;
          elsif Last > 14
          and then To_Lower(Line(1 .. 14)) = "content-type: " then
            if Last >= 23 and then To_Lower(Line(15 .. 23)) = "text/html" then
              Status := Urls.Response_Text_Html;
            else
              Must_Scan := False;
            end if;
          end if;
        elsif Result_Class = '3' then
          -- Some sort of redirect, don't scan
          Must_Scan := False; -- redirection, nothing here to scan
          Dont_Retry := True; -- Redirects aren't errors, don't try again
        end if;
        if Debug_Flag or Worth_Noting then
          Note(Integer'Image(Last) & "|" & Line(1 .. Last));
          Worth_Noting := False;
        end if;
      end loop;
      -- never get here
    end Scan_Header;
    Fetch_Header_Problem: exception;

    procedure Do_Fetch_Header(Method  : in     String) is
      use type Claw.Sockets.Error_Codes;
    begin
      if Use_Proxy then
        -- Send a proxy request. Proxies often are too stupid to use
        -- HTTP 1.1. For HTTP 1.0, they need an entire path.
        Te_Text_Line_Sockets.Put_Line(Socket,
                                      Method & " " & Page_Full_Url
                                      & " HTTP/1.0");
        Close_After := True; -- Since it's 1.0
        -- Department of redundancy department: the Wingate proxy is particularly
        -- stupid, and needs the host in the URL to find the server, but it
        -- doesn't send the host name to the server. So we repeat the Host line
        -- (that is what IE 4.0).
        -- domain is Host without initial //
        Te_Text_Line_Sockets.Put_Line(Socket,
                                      "Host: "
                                      & Host(Host'First + 2 .. Host'Last));
        Te_Text_Line_Sockets.Put_Line(Socket, "Connection: close");
        Te_Text_Line_Sockets.Put_Line(Socket, "User-agent: Ada Finder");
        -- Be nice to the server operator, and explain ourselves.
      else
        Te_Text_Line_Sockets.Put_Line(Socket,
                                      Method & " " & Path & " HTTP/1.1");
        -- domain is Host without initial //
        Te_Text_Line_Sockets.Put_Line(Socket,
                                      "Host: "
                                      & Host(Host'First + 2 .. Host'Last));
        if not Is_Root_Domain then
          Te_Text_Line_Sockets.Put_Line(Socket, "Connection: close");
        end if;
        Te_Text_Line_Sockets.Put_Line(Socket, "User-agent: Ada Finder");
      end if;
      Te_Text_Line_Sockets.Put_Line(Socket, Ascii.Cr & Ascii.Lf); -- empty line
      -- we've asked Server, now get his response
      Chunked_Transfer := False;
      -- Force to plain, in case this was a persistent socket used for chunked
      -- results previously.
      Te_Text_Line_Sockets.Set_Encoding(Socket, Te_Text_Line_Sockets.Plain);
      Content_Length_Present := False;
      Scan_Header;
    exception
      when Claw.Not_Found_Error =>
        Te_Text_Line_Sockets.Close(Socket);
        raise;
      when Argh    : others =>
        Te_Text_Line_Sockets.Get_Error_Code(Socket, Err);
        Te_Text_Line_Sockets.Close(Socket);
        if Err = Claw.Sockets.Wsae_Connreset then
          raise Claw.Not_Found_Error; -- other side has closed the socket
        end if;
        Status := Urls.Status_Codes(Err);
        Note("fh!" & Ada.Exceptions.Exception_Information(Argh));
        raise Fetch_Header_Problem;
    end Do_Fetch_Header;

    procedure Fetch_Header(Method  : in     String) is
    -- do fetch header, re-opening if closed when we try to read
    begin
      Do_Fetch_Header(Method);
    exception
      when Claw.Not_Found_Error =>
        Te_Text_Line_Sockets.Close(Socket); -- ensure closed,
        Open_Socket; -- then try once more
        if Status = Urls.No_Response_Yet then
          begin
            Do_Fetch_Header(Method);
          exception
            when Claw.Not_Found_Error =>
              Te_Text_Line_Sockets.Close(Socket);
          end;
        end if;
    end Fetch_Header;

    function Drained return Boolean is
    -- Read and toss till end or Patience seconds.  Return True iff hit end.
      use type Ada.Calendar.Time;
      Start_Time: constant Ada.Calendar.Time := Ada.Calendar.Clock;
      Patience: constant Duration := 1.0;
      Give_Up_Time: constant Ada.Calendar.Time := Start_Time + Patience;
    begin
      Urls.Operational_Statistics.Inc_Drain_Count;
      while Ada.Calendar.Clock < Give_Up_Time loop
        begin
          Te_Text_Line_Sockets.Get_Line(Socket, Line, Last, Patience);
          exit when Last < Line'First;
          Urls.Operational_Statistics.Inc_Drain_Byte_Count(
            Long_Integer(Last - Line'First + 1));
          if Last = Line'First and then Line(Line'First) = '.' then
            return True;
          end if;
        exception
          when Claw.Not_Found_Error =>
            return False;
          when others =>
            exit;
        end;
      end loop;
      Urls.Operational_Statistics.Inc_Drain_Wait(
        Ada.Calendar.Clock - Start_Time);
      return (Last = Line'First - 1);
    end Drained;
  begin
    if Debug_Flag then
      Note("debug");
    end if;
    if Is_Root_Domain then
      Urls.Operational_Statistics.Inc_Fetch_Root_Page_Count;
    end if;
    if Te_Text_Line_Sockets.Is_Open(Socket) then
      Status := Urls.No_Response_Yet;
    else
      Open_Socket; -- try to open it
      if Status /= Urls.No_Response_Yet then
        Urls.Database.Post_Work_Result(Id => Url_Id,
                                       Size => Urls.Size_Not_Available,
                                       Date => Urls.Null_Time,
                                       Status => Status,
                                       Foreign => Was_Foreign,
                                       Failed => True,
                                       Scanned => False,
                                       Url_String => Page_Full_Url,
                                       Second_Try => Is_Second_Try);
        return;
      end if;
    end if;
    Result_Class := '0'; -- Unknown.
    -- if a specific path was specified, only GET files within it
    if Urls.Root_Url_Has_Path and not Was_Foreign then
      declare
        Root_Path: constant String
          := To_Lower(Http_Urls.Path(Urls.Root_Url.all));
        Root_Directory_Last: Natural
          := Root_Path'Last - Urls.Root_Url_Filename_Length;
        Root_Directory_Length: Natural := 0;
        A_Path  : constant String := To_Lower(Path);
        A_Path_Last: Natural := A_Path'Last;
        A_Path_Length: Natural := 0;
      begin
        if Root_Path(Root_Directory_Last) = '/' then
          Root_Directory_Last := Root_Directory_Last - 1;
        end if;
        Root_Directory_Length := Root_Directory_Last - Root_Path'First + 1;
        if A_Path(A_Path_Last) = '/' then
          A_Path_Last := A_Path_Last - 1;
        end if;
        A_Path_Length := A_Path_Last - A_Path'First + 1;
        if A_Path_Length < Root_Directory_Length then
          Was_Foreign := True;
        elsif A_Path_Length = Root_Directory_Length then
          null; -- Was_Foreign stays False
        elsif not Ada.Strings.Maps.Is_In(
                    A_Path(A_Path'First + Root_Directory_Length),
                    Non_Filename_Chars) then
          Was_Foreign := True;
        elsif A_Path(A_Path'First .. A_Path'First + Root_Directory_Length - 1)
              /= Root_Path(Root_Path'First .. Root_Directory_Last) then
          Was_Foreign := True;
        end if;
        if Was_Foreign then
          Must_Scan := False; -- don't scan foreign stuff
        end if;
      end;
    end if;
    if Must_Scan then
      if (Path'Length >= 5
          and then To_Lower(Path(Path'Last - 3 .. Path'Last)) = ".htm")
      or else (Path'Length >= 6
               and then To_Lower(Path(Path'Last - 4 .. Path'Last)) = ".html")
      or else (Path'Length >= 5
               and then To_Lower(Path(Path'Last - 3 .. Path'Last)) = ".jsp")
      or else (Path'Length >= 7
               and then To_Lower(Path(Path'Last - 5 .. Path'Last)) = ".shtml")
      then
        Go_Direct_To_Get := True; -- it's an html page type file extension
      end if;
    end if;
    if not Must_Scan then
      Fetch_Header("HEAD"); -- look at header to get size, etc
      if Close_After then
        Te_Text_Line_Sockets.Close(Socket);
      end if;
    elsif Go_Direct_To_Get then
      Fetch_Header("GET");
      if not Must_Scan then
        -- We did a GET, but header indicates no need.
        -- Usually this means no connect, 404, etc.
        -- but it could be that at .htm file had
        -- content type /= text/html, in which case
        -- the server is pushing the file at us
        -- and we'll have to drain or close the socket.
        -- If the incoming page is large, draining will
        -- give up and we'll have to close the socket
        -- to turn off the incoming stream.
        if Close_After then
          Te_Text_Line_Sockets.Close(Socket);
        elsif not Drained then
          -- try to read to end, else just close
          Te_Text_Line_Sockets.Close(Socket);
        end if;
      end if;
    else
      Fetch_Header("HEAD");
      if Close_After then
        Te_Text_Line_Sockets.Close(Socket);
      end if;
    end if;
    if Must_Scan and then not Te_Text_Line_Sockets.Is_Open(Socket) then
      Open_Socket; -- try to open it
      if Status /= Urls.No_Response_Yet then
        -- can't open it
        Must_Scan := False;
        Te_Text_Line_Sockets.Close(Socket);
      end if;
    end if;
    if Must_Scan and not Go_Direct_To_Get then
      Fetch_Header("GET");
      if not Must_Scan then
        -- Weird. HEAD showed we should GET, but GET
        -- results in a header that says not to GET.
        if Close_After then
          Te_Text_Line_Sockets.Close(Socket);
        elsif not Drained then
          -- try to read to end, else just close
          Te_Text_Line_Sockets.Close(Socket);
        end if;
      end if;
    end if;
    if Must_Scan then
      -- scan it
      if Chunked_Transfer then
        Te_Text_Line_Sockets.Set_Encoding(Socket, Te_Text_Line_Sockets.Chunked);
        Content_Length_Present := False; -- SHOULD NOT be present, ignore
      else
        Te_Text_Line_Sockets.Set_Encoding(Socket, Te_Text_Line_Sockets.Plain);
      end if;
      Scanned_Item_Size := 0;
      declare
        use type Ada.Calendar.Time;
        Start_Scan_At: constant Ada.Calendar.Time := Ada.Calendar.Clock;
      begin
        Get_Links(Me, Debug_Flag, Source_Name => Page_Full_Url);
        if Ada.Calendar.Clock - Start_Scan_At > 30.0 then
          Note("scan took" & Duration'Image(Ada.Calendar.Clock - Start_Scan_At));
        end if;
        Dont_Retry := True;
        if Scanned_Item_Size > 0 then
          Item_Size := Scanned_Item_Size; -- actual observation
        end if;
        if Content_Length_Present and Content_Length_Remaining /= 0 then
          Note(Urls.Content_Lengths'Image(Content_Length_Remaining) & " unread");
        end if;
        if Chunked_Transfer then
          Te_Text_Line_Sockets.Set_Encoding(Socket, Te_Text_Line_Sockets.Plain);
          if not Drained then
            Te_Text_Line_Sockets.Close(Socket);
          end if;
        end if;
      exception
        when Te_Text_Line_Sockets.Bad_Chunk =>
          Close_After := True;
          if Ada.Calendar.Clock - Start_Scan_At > 30.0 then
            Note("scan took"
                 & Duration'Image(Ada.Calendar.Clock - Start_Scan_At));
          end if;
          Note("Bad chunk, page reading stopped");
          Status := Urls.Bad_Chunk_Read;
        when Argh : Claw.Windows_Error =>
          Close_After := True;
          if Ada.Calendar.Clock - Start_Scan_At > 30.0 then
            Note("scan took"
                 & Duration'Image(Ada.Calendar.Clock - Start_Scan_At));
          end if;
          Note("argh:" & Ada.Exceptions.Exception_Information(Argh));
          Te_Text_Line_Sockets.Get_Error_Code(Socket, Err);
          case Err is
            when Claw.Sockets.Wsae_Connreset =>
              Dont_Retry := True; -- likely host has sent all there is
            when others =>
              Status := Urls.Status_Codes(Err);
          end case;
      end;
    elsif Status = Urls.Good_200_Response
       or Status = Urls.Response_Text_Html then
      Dont_Retry := True; -- only wanted, and got, header OK
    end if;
    if Close_After then
      Te_Text_Line_Sockets.Close(Socket);
    end if;
    Urls.Database.Post_Work_Result(Id => Url_Id,
                                   Size => Item_Size,
                                   Date => Date,
                                   Status => Status,
                                   Foreign => Was_Foreign,
                                   Failed => (not Dont_Retry),
                                   Scanned => Must_Scan,
                                   Url_String => Page_Full_Url,
                                   Second_Try => Is_Second_Try);
  exception
    when Egad1 : Claw.Not_Found_Error =>
      Note("egad:" & Ada.Exceptions.Exception_Information(Egad1));
      -- likely host has sent all there is
      Urls.Database.Post_Work_Result(Id => Url_Id,
                                     Size => Item_Size,
                                     Date => Date,
                                     Status => Status,
                                     Foreign => Was_Foreign,
                                     Failed => not Dont_Retry,
                                     Scanned => False,
                                     Url_String => Page_Full_Url,
                                     Second_Try => Is_Second_Try);
      Te_Text_Line_Sockets.Close(Socket);
    when Egad2   : Claw.Windows_Error =>
      Note("egad:" & Ada.Exceptions.Exception_Information(Egad2));
      Te_Text_Line_Sockets.Get_Error_Code(Socket, Err);
      case Err is
        when Claw.Sockets.Wsae_Connreset =>
          -- likely host has sent all there is
          Urls.Database.Post_Work_Result(Id => Url_Id,
                                         Size => Item_Size,
                                         Date => Date,
                                         Status => Status,
                                         Foreign => Was_Foreign,
                                         Failed => not Dont_Retry,
                                         Scanned => False,
                                         Url_String => Page_Full_Url,
                                         Second_Try => Is_Second_Try);
        when others =>
          Status := Urls.Status_Codes(Err);
          Urls.Database.Post_Work_Result(Id => Url_Id,
                                         Size => Item_Size,
                                         Date => Date,
                                         Status => Status,
                                         Foreign => Was_Foreign,
                                         Failed => True,
                                         Scanned => Must_Scan,
                                         Url_String => Page_Full_Url,
                                         Second_Try => Is_Second_Try);
      end case;
      Te_Text_Line_Sockets.Close(Socket);
    when Fetch_Header_Problem =>
      Urls.Database.Post_Work_Result(Id => Url_Id,
                                     Size => Item_Size,
                                     Date => Date,
                                     Status => Status,
                                     Foreign => Was_Foreign,
                                     Failed => not Dont_Retry,
                                     Scanned => False,
                                     Url_String => Page_Full_Url,
                                     Second_Try => Is_Second_Try);
    when Egad : others =>
      Note("egad:" & Ada.Exceptions.Exception_Information(Egad));
      Urls.Database.Post_Work_Result(Id => Url_Id,
                                     Size => Item_Size,
                                     Date => Date,
                                     Status => Status,
                                     Foreign => Was_Foreign,
                                     Failed => True,
                                     Scanned => Must_Scan,
                                     Url_String => Page_Full_Url,
                                     Second_Try => Is_Second_Try);
      Te_Text_Line_Sockets.Close(Socket);
  end Fetch_Page;

  task body Worker_Tasks is
    use type Urls.Counts;
    Me      : Character := '?';
    Is_Second_Try: Boolean;
    Socket  : Te_Text_Line_Sockets.Socket_Type(Max_Line_Length => 5000);
    Prev_Was_Root: Boolean := False; -- just did root Host, socket may be open
    Url     : Urls.Url_Ptr_Type;
    Url_Id  : Urls.Counts; -- zero indicates no more work
    -- Root may be overwhelmed by too many simultaneous workers.
    -- To avoid that, if a worker gets a WSAE_Connrefused from Root,
    -- it will thereafter get only non-root work, thus leaving a smaller
    -- max number of workers still processing Root work.
    Dont_Use_For_Root: Boolean := False;
  begin
    select
      accept Start(Who     : in     Character) do
        Me := Who;
      end Start;
    or
      terminate;
    end select;
    loop
      Urls.Free(Url); -- toss any old stuff
      if Dont_Use_For_Root then
        -- request non-root (but may still be assigned root work)
        Urls.Database.Get_Non_Root_Work(Url_Id, Url, Is_Second_Try);
      else
        Urls.Database.Get_Any_Work(Url_Id, Url, Is_Second_Try);
      end if;
      exit when Url_Id = 0; -- no more to do
      if not Prev_Was_Root then
        -- close old non-root domain socket
        Te_Text_Line_Sockets.Close(Socket); -- OK if already closed
      end if;
      if Http_Urls.Host(Url.all) = Urls.Root_Url_Host.all then
        Prev_Was_Root := True;
      else
        Te_Text_Line_Sockets.Close(Socket); -- close whatever, non-root wanted
        Prev_Was_Root := False;
      end if;
      Fetch_Page(Me, Is_Second_Try, Url_Id,
                 Scheme => Http_Urls.Scheme(Url.all),
                 Host => Http_Urls.Host(Url.all),
                 Path => Urls.Squeeze_Path(Http_Urls.Path(Url.all)),
                 Socket => Socket,
                 Is_Root_Domain => Prev_Was_Root,
                 Root_Connrefused => Dont_Use_For_Root);
      -- Socket may still be open
    end loop;
    Urls.Free(Url);
  exception
    when others =>
      Urls.Free(Url);
  end Worker_Tasks;

end Workers;
